#include "common.h"
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "portable.h"
#include "dnautil.h"
#include "dnaseq.h"
#include "nib.h"
#include "fa.h"
#include "dystring.h"
#include "errabort.h"
#include "genoFind.h"
#include "cheapcgi.h"
#include "trans3.h"

int version = 14;
int maxSeqSize = 40000;
int maxAaSize = 8000;
int minMatch = gfMinMatch;	/** Can be overridden from command line. */
int tileSize = gfTileSize;	/** Can be overridden from command line. */
boolean doTrans = FALSE;	/** Do translation? */
boolean doTrans = FALSE;	/** Do translation? */
boolean allowOneMismatch = FALSE; 
int repMatch = 1024;
int maxGap = gfMaxGap;
FILE *logFile = NULL;
boolean seqLog = FALSE;

/** Explain usage and exit. */
void usage()
{
errAbort(
  "gfServer - Make a server to quickly find where DNA occurs in genome.\n"
  "To set up a server:\n"
  "   gfServer start host port file(s).nib\n"
  "To remove a server:\n"
  "   gfServer stop host port\n"
  "To query a server with DNA sequence:\n"
  "   gfServer query host port probe.fa\n"
  "To query a server with protein sequence:\n"
  "   gfServer protQuery host port probe.fa\n"
  "To query a server with translated dna sequence:\n"
  "   gfServer transQuery host port probe.fa\n"
  "To process one probe fa file against a .nib format genome (not starting server):\n"
  "   gfServer direct probe.fa file(s).nib\n"
  "To figure out usage level\n"
  "   gfServer status host port\n"
  "To get input file list\n"
  "   gfServer files host port\n"
  "Options:\n"
  "   -tileSize=N size of n-mers to index.  Default is 11 for nucleotides, 4 for\n"
  "               proteins (or translated nucleotides).\n"
  "   -minMatch=N Number of n-mer matches that trigger detailed alignment\n"
  "               Default is 2 for nucleotides, 3 for protiens.\n"
  "   -maxGap=N   Number of insertions or deletions allowed between n-mers.\n"
  "               Default is 2 for nucleotides, 0 for protiens.\n"
  "   -trans  Translate database to protein in 6 frames.  Note: it is best\n"
  "           to run this on RepeatMasked data in this case.\n"
  "   -log=logFile keep a log file that records server requests.\n"
  "   -seqLog    Include sequences in log file\n"
  );

}

/** Don't set up server - just directly look for matches. */
void genoFindDirect(char *probeName, int nibCount, char *nibFiles[])
{
struct genoFind *gf = NULL;
static struct genoFind *transGf[2][3];
struct lineFile *lf = lineFileOpen(probeName, TRUE);
struct dnaSeq seq;
int hitCount = 0, clumpCount = 0, oneHit;

if (doTrans)
    errAbort("Don't support translated direct stuff currently, sorry");

gf = gfIndexNibs(nibCount, nibFiles, minMatch, maxGap, tileSize, repMatch, FALSE,
	allowOneMismatch);

while (faSpeedReadNext(lf, &seq.dna, &seq.size, &seq.name))
    {
    struct lm *lm = lmInit(0);
    struct gfClump *clumpList = gfFindClumps(gf, &seq, lm, &oneHit), *clump;
    hitCount += oneHit;
    for (clump = clumpList; clump != NULL; clump = clump->next)
	{
	++clumpCount;
	printf("%s ", seq.name);
	gfClumpDump(gf, clump, stdout);
	}
    gfClumpFreeList(&clumpList);
    lmCleanup(&lm);
    }
lineFileClose(&lf);
genoFindFree(&gf);
}

/** Convert from ascii to integer. */
int getPortIx(char *portName)
{
if (!isdigit(portName[0]))
    errAbort("Expecting a port number got %s", portName);
return atoi(portName);
}
struct sockaddr_in sai;		/** Some system socket info. */
struct sockaddr_in sai;		/** Some system socket info. */

/** Set up our socket. */
int setupSocket(char *portName, char *hostName)
{
int port = getPortIx(portName);
struct hostent *hostent;
hostent = gethostbyname(hostName);
if (hostent == NULL)
    {
    errAbort("Couldn't find host %s. h_errno %d", hostName, h_errno);
    }
sai.sin_family = AF_INET;
sai.sin_port = htons(port);
memcpy(&sai.sin_addr.s_addr, hostent->h_addr_list[0], sizeof(sai.sin_addr.s_addr));
return socket(AF_INET, SOCK_STREAM, 0);
}

/** Print message to log file. */
void logIt(char *format, ...)
{
va_list args;
if (logFile != NULL)
    {
    fprintf(logFile, "%lu ", clock1000());
    va_start(args, format);
    vfprintf(logFile, format, args);
    va_end(args);
    fflush(logFile);
    }
}

/** Some variables to gather statistics on usage. */
long baseCount = 0, queryCount = 0, aaCount = 0;
int warnCount = 0;
int noSigCount = 0;
int missCount = 0;
int trimCount = 0;

void dnaQuery(struct genoFind *gf, struct dnaSeq *seq, 
	int connectionHandle, char buf[256])	
/** Handle a query for DNA/DNA match. */
{
struct gfClump *clumpList = NULL, *clump;
int limit = 100;
int clumpCount = 0, hitCount = -1;
struct lm *lm = lmInit(0);

clumpList = gfFindClumps(gf, seq, lm, &hitCount);
if (clumpList == NULL)
    ++missCount;
for (clump = clumpList; clump != NULL; clump = clump->next)
    {
    struct gfSeqSource *ss = clump->target;
    sprintf(buf, "%d\t%d\t%s\t%d\t%d\t%d", 
	clump->qStart, clump->qEnd, ss->fileName,
	clump->tStart-ss->start, clump->tEnd-ss->start, clump->hitCount);
    gfSendString(connectionHandle, buf);
    ++clumpCount;
    if (--limit < 0)
	break;
    }
gfClumpFreeList(&clumpList);
lmCleanup(&lm);
logIt("%d clumps, %d hits\n", clumpCount, hitCount);
}

void transQuery(struct genoFind *transGf[2][3], aaSeq *seq, 
	int connectionHandle, char buf[256])	
/** Handle a query for protein/translated DNA match. */
{
struct gfClump *clumps[3], *clump;
int isRc, frame;
char strand;
struct dyString *dy  = newDyString(1024);
struct gfHit *hit;
int clumpCount = 0, hitCount = 0, oneHit;
struct lm *lm = lmInit(0);

sprintf(buf, "tileSize %d", tileSize);
gfSendString(connectionHandle, buf);
for (frame = 0; frame < 3; ++frame)
    clumps[frame] = NULL;
for (isRc = 0; isRc <= 1; ++isRc)
    {
    strand = (isRc ? '-' : '+');
    gfTransFindClumps(transGf[isRc], seq, clumps, lm, &oneHit);
    hitCount += oneHit;
    for (frame = 0; frame < 3; ++frame)
        {
	int limit = 200;
	for (clump = clumps[frame]; clump != NULL; clump = clump->next)
	    {
	    struct gfSeqSource *ss = clump->target;
	    sprintf(buf, "%d\t%d\t%s\t%d\t%d\t%d\t%c\t%d", 
		clump->qStart, clump->qEnd, ss->fileName,
		clump->tStart-ss->start, clump->tEnd-ss->start, clump->hitCount,
		strand, frame);
	    gfSendString(connectionHandle, buf);
	    dyStringClear(dy);
	    for (hit = clump->hitList; hit != NULL; hit = hit->next)
	        dyStringPrintf(dy, " %d %d", hit->qStart, hit->tStart - ss->start);
	    gfSendLongString(connectionHandle, dy->string);
	    ++clumpCount;
	    if (--limit < 0)
		break;
	    }
	gfClumpFreeList(&clumps[frame]);
	}
    }
if (clumpCount == 0)
    ++missCount;
freeDyString(&dy);
lmCleanup(&lm);
logIt("%d clumps, %d hits\n", clumpCount, hitCount);
}

void transTransQuery(struct genoFind *transGf[2][3], struct dnaSeq *seq, 
	int connectionHandle, char buf[256])	
/** Handle a query for protein/translated DNA match. */
{
struct gfClump *clumps[3][3], *clump;
int isRc, qFrame, tFrame;
char strand;
struct trans3 *t3 = trans3New(seq);
struct dyString *dy  = newDyString(1024);
struct gfHit *hit;
int clumpCount = 0, hitCount = 0, oneCount;

sprintf(buf, "tileSize %d", tileSize);
gfSendString(connectionHandle, buf);
for (qFrame = 0; qFrame<3; ++qFrame)
    for (tFrame=0; tFrame<3; ++tFrame)
	clumps[qFrame][tFrame] = NULL;
for (isRc = 0; isRc <= 1; ++isRc)
    {
    struct lm *lm = lmInit(0);
    strand = (isRc ? '-' : '+');
    gfTransTransFindClumps(transGf[isRc], t3->trans, clumps, lm, &oneCount);
    hitCount += oneCount;
    for (qFrame = 0; qFrame<3; ++qFrame)
	{
	for (tFrame=0; tFrame<3; ++tFrame)
	    {
	    int limit = 200;
	    for (clump = clumps[qFrame][tFrame]; clump != NULL; clump = clump->next)
		{
		struct gfSeqSource *ss = clump->target;
		sprintf(buf, "%d\t%d\t%s\t%d\t%d\t%d\t%c\t%d\t%d", 
		    clump->qStart, clump->qEnd, ss->fileName,
		    clump->tStart-ss->start, clump->tEnd-ss->start, clump->hitCount,
		    strand, qFrame, tFrame);
		gfSendString(connectionHandle, buf);
		dyStringClear(dy);
		for (hit = clump->hitList; hit != NULL; hit = hit->next)
		    {
		    dyStringPrintf(dy, " %d %d", hit->qStart, hit->tStart - ss->start);
		    }
		gfSendLongString(connectionHandle, dy->string);
		++clumpCount;
		if (--limit < 0)
		    break;
		}
	    gfClumpFreeList(&clumps[qFrame][tFrame]);
	    }
	}
    lmCleanup(&lm);
    }
trans3Free(&t3);
if (clumpCount == 0)
    ++missCount;
logIt("%d clumps, %d hits\n", clumpCount, hitCount);
}

/** Load up index and hang out in RAM. */
void startServer(char *hostName, char *portName, int nibCount, char *nibFiles[])
{
struct genoFind *gf = NULL;
static struct genoFind *transGf[2][3];
char buf[256];
char *line, *command;
int fromLen, readSize, res;
int socketHandle = 0, connectionHandle = 0;
char *logFileName = cgiOptionalString("log");
signal(SIGPIPE, SIG_IGN);	/** Block broken pipe signals. */
signal(SIGPIPE, SIG_IGN);	/** Block broken pipe signals. */
if (logFileName != NULL)
    logFile = mustOpen(logFileName, "a");
logIt("gfServer version %d on host %s, port %s\n", version, hostName, portName);
if (doTrans)
    {
    logIt("setting up translated index\n");
    gfIndexTransNibs(transGf, nibCount, nibFiles, 
    	minMatch, maxGap, tileSize, repMatch, NULL, allowOneMismatch);
    }
else
    {
    gf = gfIndexNibs(nibCount, nibFiles, minMatch, 
    	maxGap, tileSize, repMatch, NULL, allowOneMismatch);
    }
logIt("indexing complete\n");

/** Set up socket.  Get ready to listen to it. */
socketHandle = setupSocket(portName, hostName);
if (bind(socketHandle, (struct sockaddr*)&sai, sizeof(sai)) == -1)
     errAbort("Couldn't bind to %s port %s", hostName, portName);
res = listen(socketHandle, 100);

logIt("Server ready for queries!\n");
printf("Server ready for queries!\n");
for (;;)
    {
    connectionHandle = accept(socketHandle, NULL, &fromLen);
    readSize = read(connectionHandle, buf, sizeof(buf)-1);
    buf[readSize] = 0;
    logIt("%s\n", buf);
    if (!startsWith(gfSignature(), buf))
        {
	++noSigCount;
	continue;
	}
    line = buf + strlen(gfSignature());
    command = nextWord(&line);
    if (sameString("quit", command))
        {
	logIt("Ignoring quit message\n");
	break;
	}
    else if (sameString("status", command))
        {
	sprintf(buf, "version %d", version);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "type %s", (doTrans ? "translated" : "nucleotide"));
	gfSendString(connectionHandle, buf);
	sprintf(buf, "host %s", hostName);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "port %s", portName);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "tileSize %d", tileSize);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "minMatch %d", minMatch);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "requests %ld", queryCount);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "bases %ld", baseCount);
	gfSendString(connectionHandle, buf);
	if (doTrans)
	    {
	    sprintf(buf, "aa %ld", aaCount);
	    gfSendString(connectionHandle, buf);
	    }
	sprintf(buf, "misses %d", missCount);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "noSig %d", noSigCount);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "trimmed %d", trimCount);
	gfSendString(connectionHandle, buf);
	sprintf(buf, "warnings %d", warnCount);
	gfSendString(connectionHandle, buf);
	gfSendString(connectionHandle, "end");
	}
    else if (sameString("query", command) || 
    	sameString("protQuery", command) || sameString("transQuery", command))
        {
	int querySize;
	boolean queryIsProt = sameString(command, "protQuery");
	char *s = nextWord(&line);
	if (s == NULL || !isdigit(s[0]))
	    {
	    warn("Expecting query size after query command");
	    ++warnCount;
	    }
	else
	    {
	    struct dnaSeq seq;
	    char *type = NULL;

	    if (queryIsProt && !doTrans)
	        {
		warn("protein query sent to nucleotide server");
		++warnCount;
		queryIsProt = FALSE;
		}
	    buf[0] = 'Y';
	    if (write(connectionHandle, buf, 1) == 1)
		{
		seq.size = atoi(s);
		seq.name = NULL;
		if (seq.size > 0)
		    {
		    ++queryCount;
		    seq.dna = needLargeMem(seq.size+1);
		    if (gfReadMulti(connectionHandle, seq.dna, seq.size) != seq.size)
			{
			warn("Didn't sockRecieveString all %d bytes of query sequence", seq.size);
			++warnCount;
			}
		    else
			{
			int maxSize = (doTrans ? maxAaSize : maxSeqSize);

			seq.dna[seq.size] = 0;
			if (queryIsProt)
			    {
			    seq.size = aaFilteredSize(seq.dna);
			    aaFilter(seq.dna, seq.dna);
			    }
			else
			    {
			    seq.size = dnaFilteredSize(seq.dna);
			    dnaFilter(seq.dna, seq.dna);
			    }
			if (seq.size > maxSize)
			    {
			    ++trimCount;
			    seq.size = maxSize;
			    seq.dna[maxSize] = 0;
			    }
			if (queryIsProt)
			    aaCount += seq.size;
			else
			    baseCount += seq.size;
			if (seqLog && logFile != NULL)
			    {
			    faWriteNext(logFile, "query", seq.dna, seq.size);
			    fflush(logFile);
			    }
			if (doTrans)
			   {
			   if (queryIsProt)
				transQuery(transGf, &seq, connectionHandle, buf);
			   else
				transTransQuery(transGf, &seq, 
				    connectionHandle, buf);
			   }
			else
			    dnaQuery(gf, &seq, connectionHandle, buf);
			}
		    freez(&seq.dna);
		    }
		gfSendString(connectionHandle, "end");
		}
	    }
	}
    else if (sameString("files", command))
        {
	struct gfSeqSource *ss;
	int i;
	sprintf(buf, "%d", nibCount);
	gfSendString(connectionHandle, buf);
	for (i=0; i<nibCount; ++i)
	    {
	    sprintf(buf, "%s", nibFiles[i]);
	    gfSendString(connectionHandle, buf);
	    }
	}
    else
        {
	warn("Unknown command %s", command);
	++warnCount;
	}
    close(connectionHandle);
    connectionHandle = 0;
    }
close(socketHandle);
}

/** Send stop message to server. */
void stopServer(char *hostName, char *portName)
{
char buf[256];
int sd = 0;


/** Set up socket.  Get ready to listen to it. */
sd = setupSocket(portName, hostName);
if (connect(sd, (struct sockaddr*)&sai, sizeof(sai)) == -1)
     errAbort("Couldn't connect to %s port %s", hostName, portName);

/** Put together quit command. */
sprintf(buf, "%squit", gfSignature());
write(sd, buf, strlen(buf));

close(sd);
printf("sent stop message to server\n");
}

/** Send status message to server arnd report result. */
void statusServer(char *hostName, char *portName)
{
char buf[256];
char *line, *command;
int fromLen, readSize;
int sd = 0;
int fileCount;
int i;

/** Set up socket.  Get ready to listen to it. */
sd = setupSocket(portName, hostName);
if (connect(sd, (struct sockaddr*)&sai, sizeof(sai)) == -1)
     errAbort("Couldn't connect to %s port %s", hostName, portName);

/** Put together command. */
sprintf(buf, "%sstatus", gfSignature());
write(sd, buf, strlen(buf));

for (;;)
    {
    if (gfGetString(sd, buf) == NULL)
        break;
    if (sameString(buf, "end"))
        break;
    else
        printf("%s\n", buf);
    }
close(sd);
}

void queryServer(char *type, 
	char *hostName, char *portName, char *faName, boolean complex, boolean isProt)
/** Send simple query to server and report results. */
{
char buf[256];
char *line, *command;
int fromLen, readSize;
int sd = 0;
bioSeq *seq = faReadSeq(faName, !isProt);
int matchCount = 0;


/** Set up socket.  Get ready to listen to it. */
sd = setupSocket(portName, hostName);
if (connect(sd, (struct sockaddr*)&sai, sizeof(sai)) == -1)
     errAbort("Couldn't connect to %s port %s", hostName, portName);

/** Put together query command. */
sprintf(buf, "%s%s %d", gfSignature(), type, seq->size);
write(sd, buf, strlen(buf));

read(sd, buf, 1);
if (buf[0] != 'Y')
    errAbort("Expecting 'Y' from server, got %c", buf[0]);
write(sd, seq->dna, seq->size);

if (complex)
    {
    char *s = gfRecieveString(sd, buf);
    printf("%s\n", s);
    }

for (;;)
    {
    if (gfGetString(sd, buf) == NULL)
        break;
    if (sameString(buf, "end"))
	{
	printf("%d matches\n", matchCount);
	break;
	}
    else
	{
        printf("%s\n", buf);
	if (complex)
	    {
	    char *s = gfGetLongString(sd);
	    if (s == NULL)
	        break;
	    printf("%s\n", s);
	    freeMem(s);
	    }
	}
    ++matchCount;
    }
close(sd);
}

/** Get and display input file list. */
void getFileList(char *hostName, char *portName)
{
char buf[256];
char *line, *command;
int fromLen, readSize;
int sd = 0;
int fileCount;
int i;

/** Set up socket.  Get ready to listen to it. */
sd = setupSocket(portName, hostName);
if (connect(sd, (struct sockaddr*)&sai, sizeof(sai)) == -1)
     errAbort("Couldn't connect to %s port %s", hostName, portName);

/** Put together command. */
sprintf(buf, "%sfiles", gfSignature());
write(sd, buf, strlen(buf));

/** Get count of files, and then each file name. */
if (gfGetString(sd, buf) != NULL)
    {
    fileCount = atoi(buf);
    for (i=0; i<fileCount; ++i)
	{
	printf("%s\n", gfRecieveString(sd, buf));
	}
    }
close(sd);
}

/** Process command line. */
int main(int argc, char *argv[])
{
char *command;

gfCatchPipes();
cgiSpoof(&argc, argv);
command = argv[1];
if (cgiBoolean("trans"))
    {
    doTrans = TRUE;
    tileSize = 4;
    minMatch = 3;
    maxGap = 0;
    repMatch = gfPepMaxTileUse;
    }
tileSize = cgiOptionalInt("tileSize", tileSize);
minMatch = cgiOptionalInt("minMatch", minMatch);
repMatch = cgiOptionalInt("repMatch", repMatch);
seqLog = cgiBoolean("seqLog");
if (argc < 2)
    usage();
if (sameWord(command, "direct"))
    {
    if (argc < 4)
        usage();
    genoFindDirect(argv[2], argc-3, argv+3);
    }
else if (sameWord(command, "start"))
    {
    if (argc < 5)
        usage();
    startServer(argv[2], argv[3], argc-4, argv+4);
    }
else if (sameWord(command, "stop"))
    {
    if (argc != 4)
	usage();
    stopServer(argv[2], argv[3]);
    }
else if (sameWord(command, "query"))
    {
    if (argc != 5)
	usage();
    queryServer(command, argv[2], argv[3], argv[4], FALSE, FALSE);
    }
else if (sameWord(command, "protQuery"))
    {
    if (argc != 5)
	usage();
    queryServer(command, argv[2], argv[3], argv[4], TRUE, TRUE);
    }
else if (sameWord(command, "transQuery"))
    {
    if (argc != 5)
	usage();
    queryServer(command, argv[2], argv[3], argv[4], TRUE, FALSE);
    }
else if (sameWord(command, "status"))
    {
    if (argc != 4)
	usage();
    statusServer(argv[2], argv[3]);
    }
else if (sameWord(command, "files"))
    {
    if (argc != 4)
	usage();
    getFileList(argv[2], argv[3]);
    }
else
    {
    usage();
    }
return 0;
}
