/******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/** jksql.h - Stuff to manage interface with SQL database. 
 *
 * To use - first open a connection, then pass a SQL query to 
 * sqlGetResult, then use sqlNextRow to examine result row by
 * row. The returned row is just an array of strings.  Use
 * sqlUnsigned, sqlSigned, and atof to convert numeric results
 * to normal form. 
 *
 * These routines will all print an error message and abort if 
 * there's a problem, cleaning up open connections, etc. on abort 
 * (or on program exit).  Do a pushAbortHandler if you want to 
 * catch the aborts.  The error messages from bad SQL syntax
 * are actually pretty good (they're just passed on from
 * mySQL). */

#ifndef SQLNUM_H
#include "sqlNum.h"
#endif

#ifndef SQLLIST_H
#include "sqlList.h"
#endif

	
/** Connect to database on default host as default user. */
struct sqlConnection *sqlConnect(char *database);

struct sqlConnection *sqlConnectRemote(char *host, 
	char *user, char *password, char *database);
/** Connect to database somewhere as somebody. */

/** Close down connection. */
void sqlDisconnect(struct sqlConnection **pSc);

/** Return a new connection cache. (Useful if going to be
 * doing lots of different queries in different routines
 * to same database - reduces connection overhead.) */
struct sqlConnCache *sqlNewConnCache(char *database);

/** Dispose of a connection cache. */
void sqlFreeConnCache(struct sqlConnCache **pCache);

/** Allocate a cached connection. */
struct sqlConnection *sqlAllocConnection(struct sqlConnCache *cache);

/** Free up a cached connection. */
void sqlFreeConnection(struct sqlConnCache *cache,struct sqlConnection **pConn);

/** Tell database to do something that produces no results table. */
void sqlUpdate(struct sqlConnection *conn, char *query);

/** Query database and return TRUE if it had a non-empty result. */
boolean sqlExists(struct sqlConnection *conn, char *query);

/** Query database.
 * Returns NULL if result was empty.  Otherwise returns a structure
 * that you can do sqlRow() on. */
struct sqlResult *sqlGetResult(struct sqlConnection *sc, char *query);

/** Query database. If result empty squawk and die. */
struct sqlResult *sqlMustGetResult(struct sqlConnection *sc, char *query);

/** Free up a result. */
void sqlFreeResult(struct sqlResult **pRes);

/** Count the number of columns in result. */
int sqlCountColumns(struct sqlResult *sr);

/** Return TRUE if a table exists. */
boolean sqlTableExists(struct sqlConnection *sc, char *table);

/** Does query and returns first field in first row.  Meant
 * for cases where you are just looking up one small thing.  
 * Returns NULL if query comes up empty. */
char *sqlQuickQuery(struct sqlConnection *sc, char *query, char *buf, int bufSize);

/** Get numerical result from simple query */
int sqlQuickNum(struct sqlConnection *conn, char *query);

/** Create table from query if it doesn't exist already. 
 * Returns FALSE if didn't make table. */
boolean sqlMaybeMakeTable(struct sqlConnection *sc, char *table, char *query);

/** Fetch next row from result.  If you need to save these strings
 * past the next call to sqlNextRow you must copy them elsewhere.
 * It is ok to write to the strings - replacing tabs with zeroes
 * for instance.  You can call this with a NULL sqlResult.  It
 * will then return a NULL row. */
char **sqlNextRow(struct sqlResult *sr);

/** repeated calls to this function returns the names of the fields 
 * the given result */
char* sqlFieldName(struct sqlResult *sr);

/** Find number of rows in table. */
int sqlTableSize(struct sqlConnection *conn, char *table);

/** Error message handler. */
void sqlVaWarn(struct sqlConnection *sc, char *format, va_list args);

/** Printf formatted error message that adds on sql 
 * error message. */
void sqlWarn(struct sqlConnection *sc, char *format, ...);

/** Printf formatted error message that adds on sql 
 * error message and abort. */
void sqlAbort(struct sqlConnection  *sc, char *format, ...);

/** Cleanup all open connections and resources. */
void sqlCleanupAll();
