/*** diGraph - Directed graph routines. */
#ifndef DIGRAPH_H
#define DIGRAPH_H

#ifndef DLIST_H
#include "dlist.h"
#endif

/*** A directed graph. */
struct diGraph
    {
    struct dgNode *nodeList;
    struct hash *nodeHash;
    struct dlList *edgeList;
    };

/*** A node on a directed graph */
    struct dgNode *next;   /*** Next in master node list. */
    char *name;            /*** Name of node (allocated in hash). */
    void *val;             /*** Value of node. */
    struct dgConnection *nextList;  /*** Ways out of node. */
    struct dgConnection *prevList;  /*** Ways into node. */
    int topoOrder;      /*** Topographical order after call to dgTopoSort. */
    int component;      /*** Connected component number after call to dgConnectedComponents. */
    int priority;       /*** Node priority. */
    struct dgNode *tempEntry;  /*** Way in for breadth first search */
    bool visited;       /*** Visit order in depth first traversal. */
    bool conn;	        /*** Flag for connection algorithms. */
    bool flagA;         /*** Flag used in rangeGraph algorithms. */
    bool flagB;         /*** Currently unused flag. */
    bool flagB;         /*** Currently unused flag. */
    bool flagB;         /*** Currently unused flag. */
    };

/*** An edge on graph */
    void *val;		/*** Value of edge. */
    struct dgNode *a;   /*** Node on incoming side. */
    struct dgNode *b;   /*** Node on outgoing side. */
    struct dgNode *b;   /*** Node on outgoing side. */
    struct dgNode *b;   /*** Node on outgoing side. */
    };

/*** Connects one node to another. Linked to edge. */
    struct dgConnection *next;  /*** Next in list. */
    struct dgNode *node;        /*** Node this connects to. */
    struct dlNode *edgeOnList;  /*** Associated edge. */
    struct dlNode *edgeOnList;  /*** Associated edge. */
    struct dlNode *edgeOnList;  /*** Associated edge. */
    };

/*** A reference to a node. */
    struct dgNodeRef *next;	/*** Next in list. */
    struct dgNode *node;        /*** Pointer to node. */
    struct dgNode *node;        /*** Pointer to node. */
    struct dgNode *node;        /*** Pointer to node. */
    };

/*** A reference to an edge. */
    struct dgEdgeRef *next;	/*** Next in list. */
    struct dgEdge *edge;        /*** Reference to edge. */
    struct dgEdge *edge;        /*** Reference to edge. */
    struct dgEdge *edge;        /*** Reference to edge. */
    };

/*** Return a new directed graph object. */
struct diGraph *dgNew();

/*** Free a directed graph. */
void dgFree(struct diGraph **pGraph);

/*** Create new node in graph. */
struct dgNode *dgAddNode(struct diGraph *dg, char *name, void *val);

/*** Create new node with a number instead of a name. */
struct dgNode *dgAddNumberedNode(struct diGraph *dg, int id, void *val);
 
/*** Find existing node in graph. */
struct dgNode *dgFindNode(struct diGraph *dg, char *name);

/*** Find node given number. */
struct dgNode *dgFindNumberedNode(struct diGraph *dg, int id);

/*** Return value associated with node. */
void *dgNodeVal(struct dgNode *node);

/*** Return name associated with node. */
void *dgNodeName(struct dgNode *node);

/*** Return number of node.  (Will likely return 0 if node
 * was added with a name rather than a number). */
int dgNodeNumber(struct dgNode *node);

/*** Connect node a to node b.  Returns connecting edge. */
struct dgEdge *dgConnect(struct diGraph *dg, struct dgNode *a, struct dgNode *b);

/*** Connect node a to node b and put val on edge. */
struct dgEdge *dgConnectWithVal(struct diGraph *dg, struct dgNode *a, struct dgNode *b, void *val);

/*** Disconnect nodes a and b. */
void dgDisconnect(struct diGraph *dg, struct dgNode *a, struct dgNode *b);

/*** Return TRUE if b directly follows a. */
struct dgEdge *dgDirectlyFollows(struct diGraph *dg, struct dgNode *a, struct dgNode *b);

/*** Return TRUE if there's a path from a to b. */
boolean dgPathExists(struct diGraph *dg, struct dgNode *a, struct dgNode *b);

/*** Find shortest path from a to b.  Return NULL if can't be found. */
struct dgNodeRef *dgFindPath(struct diGraph *dg, struct dgNode *a, struct dgNode *b);

/*** Return reference list of all nodes connected to a, including a.
 * slFreeList this list when done. */
struct dgNodeRef *dgFindConnectedToNode(struct diGraph *dg, struct dgNode *a);

/*** Find a connected component guaranteed not to be covered before 
 * that includes a. */
struct dgNodeRef *dgFindNewConnected(struct diGraph *dg, struct dgNode *a);

/*** Find a connected component guaranteed not to be covered before 
 * that includes a.  Connected components must have values*/
struct dgNodeRef *dgFindNewConnectedWithVals(struct diGraph *dg, struct dgNode *a);

/*** Return list of nodes that make up next connected component,
 * or NULL if no more components.  slFreeList this when
 * done.  Call "dgClearConnFlags" before first call to this.
 * Do not call dgFindConnectedToNode between dgFindFirstConnected 
 * and this as they use the same flag variables to keep track of
 * what vertices are used. */
struct dgNodeRef *dgFindNextConnected(struct diGraph *dg);

/*** Like dgFindConnected, but only considers graph nodes that
 * have a val attatched. */
struct dgNodeRef *dgFindNextConnectedWithVals(struct diGraph *dg);

/*** Count number of connected components and set component field
 * of each node to reflect which component it is in. */
int dgConnectedComponents(struct diGraph *dg);

/*** Count number of connected components and set component field
 * of each node to reflect which component it is in. Only
 * consider components with values. */
int dgConnectedComponentsWithVals(struct diGraph *dg);

/*** Fill in topological order of nodes. */
void dgTopoSort(struct diGraph *dg);

/*** Return TRUE if directed graph has cycles. */
boolean dgHasCycles(struct diGraph *dg);

/*** Return TRUE if all parents of node have  been visited. */
boolean dgParentsAllVisited(struct dgNode *node);

/*** Return traversal of graph in priority order subject to
 * constraint that all parents must be output before
 * their children regardless of node priority. 
 * Graph must be cycle free. */
struct dgNodeRef *dgConstrainedPriorityOrder(struct diGraph *dg);

boolean dgRangesConsistent(struct diGraph *rangeGraph,
/*** Return TRUE if graph with a range of allowable distances associated
 * with each edge is internally consistent. */
   boolean (*findEdgeRange)(struct dgEdge *edge, int *retMin, int *retMax) );

boolean dgAddedRangeConsistent(struct diGraph *rangeGraph,
   struct dgNode *a, struct dgNode *b, int abMin, int abMax,
/*** Return TRUE if graph with a range of allowable distances associated
 * with each edge would be internally consistent if add edge from a to b
 * with given min/max values. */
   boolean (*findEdgeRange)(struct dgEdge *edge, int *retMin, int *retMax) );

/*** Return list of edges in graph that connected together nodes in subGraph. */
struct dgEdgeRef *dgFindSubEdges(struct diGraph *dg, struct dgNodeRef *subGraph);

/*** Swap polarity of all edges in erList.  (Assumes you don't have
 * edges going both directions in graph.) */
void dgSwapEdges(struct diGraph *dg, struct dgEdgeRef *erList);

/*** Clear out conn flags. */
void dgClearConnFlags(struct diGraph *dg);

/*** Clear out visit order flags. */
void dgClearVisitFlags(struct diGraph *dg);

/*** Return reference to node if in list, or NULL if not. */
struct dgNodeRef *dgFindNodeInRefList(struct dgNodeRef *refList, struct dgNode *node);

/*** Return connection to node if in list, or NULL if not. */
struct dgConnection *dgFindNodeInConList(struct dgConnection *conList, struct dgNode *node);

/*** Return list of connections out of this node. Don't free (or rearrange)
 * this list please. */
struct dgConnection *dgNextList(struct dgNode *node);

/*** Return list of connections into this node. Don't free (or rearrange)
 * this list please. */
struct dgConnection *dgPrevList(struct dgNode *node);

/*** Dump info on graph to output. */
void dgDumpGraph(struct diGraph *dg, FILE *out, boolean hideIsolated);

#endif /*** DIGRAPH_H */

