/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** Memgfx - stuff to do graphics in memory buffers.
 * Typically will just write these out as .gif files.
 * This stuff is byte-a-pixel for simplicity.
 * It can do 256 colors.
 */
#ifndef MEMGFX_H
#define MEMGFX_H

#define MG_WHITE 0
#define MG_BLACK 1
#define MG_RED 2
#define MG_GREEN 3
#define MG_BLUE 4
#define MG_CYAN 5
#define MG_MAGENTA 6
#define MG_YELLOW 7
#define MG_GRAY 8
#define MG_FREE_COLORS_START 9


typedef unsigned char Color;

struct rgbColor
    {
    unsigned char r, g, b;
    };

struct memGfx
    {
    Color *pixels;
    int width, height;
    struct rgbColor colorMap[256];
    int colorsUsed;
    int clipMinX, clipMaxX;
    int clipMinY, clipMaxY;
    };

/*** Get a new thing to draw on in memory. */
struct memGfx *mgNew(int width, int height);

/*** Free up memory raster. */
void mgFree(struct memGfx **pmg);

/*** Set all pixels to background. */
void mgClearPixels(struct memGfx *mg);

/*** Set clipping rectangle. */
void mgSetClip(struct memGfx *mg, int x, int y, int width, int height);

/*** Set clipping rect cover full thing. */
void mgUnclip(struct memGfx *mg);

/*** Returns closest color in color map to rgb values.  If it doesn't
 * already exist in color map and there's room, it will create
 * exact color in map. */
Color mgFindColor(struct memGfx *mg, unsigned char r, unsigned char g, unsigned char b);

/*** Returns closest color in color map to r,g,b */
Color mgClosestColor(struct memGfx *mg, unsigned char r, unsigned char g, unsigned char b);

/*** Adds color to end of color map if there's room. */
Color mgAddColor(struct memGfx *mg, unsigned char r, unsigned char g, unsigned char b);

/*** Returns # of unused colors in color map. */
int mgColorsFree(struct memGfx *mg);


/*** Get what to add to get to next line */
#define _mgBpr(mg) ((mg)->width)

/*** Get pixel address */
#define _mgPixAdr(mg,x,y) ((mg)->pixels+_mgBpr(mg) * (y) + (x))

/*** Unclipped plot a dot */
#define _mgPutDot(mg, x, y, color) (*_mgPixAdr(mg,x,y) = (color))

/*** Clipped put dot */
#define mgPutDot(mg,x,y,color) if ((x)>=(mg)->clipMinX && (x) < (mg)->clipMaxX && (y)>=(mg)->clipMinY  && (y) < (mg)->clipMaxY) _mgPutDot(mg,x,y,color)

/*** Put a series of dots starting at x, y and going to right width pixels. */
void mgPutSeg(struct memGfx *mg, int x, int y, int width, Color *dots);

/*** Put a series of dots starting at x, y and going to right width pixels.
 * Don't put zero dots though. */
void mgPutSegZeroClear(struct memGfx *mg, int x, int y, int width, Color *dots);

/*** Draw a (horizontal) box */
void mgDrawBox(struct memGfx *mg, int x, int y, int width, int height, Color color);


/*** Draw a line from one point to another. */
void mgDrawLine(struct memGfx *mg, int x1, int y1, int x2, int y2, Color color);

/*** Save memory bitmap as a gif. */
void mgSaveGif(struct memGfx *mg, char *name);

/*** Save GIF to an already open file. */
boolean mgSaveToGif(FILE *gif_file, struct memGfx *screen);


typedef void (*TextBlit)(int bitWidth, int bitHeight, int bitX, int bitY,
	unsigned char *bitData, int bitDataRowBytes, 
	struct memGfx *dest, int destX, int destY, 
	Color color, Color backgroundColor);
/*** This defines the type of a function that takes a rectangular
 * area of a bitplane and expands it into a rectangular area
 * of a full color screen. */

void mgTextBlit(int bitWidth, int bitHeight, int bitX, int bitY,
	unsigned char *bitData, int bitDataRowBytes, 
	struct memGfx *dest, int destX, int destY, 
	Color color, Color backgroundColor);
/*** This function leaves the background as it was. */

void mgTextBlitSolid(int bitWidth, int bitHeight, int bitX, int bitY,
	unsigned char *bitData, int bitDataRowBytes, 
	struct memGfx *dest, int destX, int destY, 
	Color color, Color backgroundColor);
/*** This function sets the background to the background color. */

typedef struct font_hdr MgFont;
/*** Type of our font.  */

MgFont *mgTinyFont();
MgFont *mgSmallFont();
MgFont *mgSmallishFont();
MgFont *mgMediumFont();
MgFont *mgLargeFont();
MgFont *mgHugeFont();

void mgText(struct memGfx *mg, int x, int y, Color color, 
	MgFont *font, char *text);
/*** Draw a line of text with upper left corner x,y. */

void mgTextCentered(struct memGfx *mg, int x, int y, int width, int height, 
	Color color, MgFont *font, char *text);
/*** Draw a line of text centered in box defined by x/y/width/height */

void mgTextRight(struct memGfx *mg, int x, int y, int width, int height, 
	Color color, MgFont *font, char *text);
/*** Draw a line of text right justified in box defined by x/y/width/height */

/*** How high in pixels is font? */
int mgFontPixelHeight(MgFont *font);

/*** How many pixels to next line ideally? */
int mgFontLineHeight(MgFont *font);

/*** How wide are a couple of letters? */
int mgFontWidth(MgFont *font, char *chars, int charCount);

/*** How wide is a string? */
int mgFontStringWidth(MgFont *font, char *string);

/*** How wide is a character? */
int mgFontCharWidth(MgFont *font, char c);


void mgDrawRuler(struct memGfx *mg, int xOff, int yOff, int height, int width,
        Color color, MgFont *font,
        int startNum, int range);
/*** Draw a ruler inside the indicated part of mg with numbers that start at
 * startNum and span range.  */

void mgDrawRulerBumpText(struct memGfx *mg, int xOff, int yOff, 
	int height, int width,
        Color color, MgFont *font,
        int startNum, int range, int bumpX, int bumpY);
/*** Draw a ruler inside the indicated part of mg with numbers that start at
 * startNum and span range.  Bump text positions slightly. */

void mgBarbedHorizontalLine(struct memGfx *mg, int x, int y, 
	int width, int barbHeight, int barbSpacing, int barbDir, Color color);
/*** Draw a horizontal line starting at xOff, yOff of given width.  Will
 * put barbs (successive arrowheads) to indicate direction of line.  
 * BarbDir of 1 points barbs to right, of -1 points them to left. */

#endif /*** MEMGFX_H */
