/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** WormDNA - stuff that finds C. elegans sequence data. */
#ifndef WORMDNA_H
#define WORMDNA_H

#ifndef DNAUTIL_H
#include "dnautil.h"
#endif

#ifndef DNASEQ_H
#include "dnaseq.h"
#endif 

#ifndef NT4_H
#include "nt4.h"
#endif 

#ifndef GDF_H
#include "gdf.h"
#endif

/*** Extra info stored in cDNA database other than string. */
    char *motherString;         /*** Holds memory for other strings. */
    char *name;                 /*** Name of cDNA. */
    char *gene;                 /*** Something like unc-1 */
    char orientation;           /*** + or - (relative to gene transcription direction) */
    char *product;              /*** Something like "cyclin-dependent kinase" */
    int cdsStart, cdsEnd;       /*** Start and stop of coding region within cDNA. */
    boolean knowStart, knowEnd; /*** Start known?  End known? */
    boolean isEmbryonic;        /*** True if derived from embryo culture. */
    boolean isAdult;            /*** True if derived from adult culture. */
    boolean isMale;             /*** True if males only. */
    boolean isHermaphrodite;    /*** True if hermaphrodites only. */
    char *description;          /*** One line description. */
    char *description;          /*** One line description. */
    char *description;          /*** One line description. */
    };

/*** Get info about cDNA sequence. */
boolean wormCdnaInfo(char *name, struct wormCdnaInfo *retInfo);

/*** Process line from .fa file containing information about cDNA into binary
 * structure. */
void wormFaCommentIntoInfo(char *faComment, struct wormCdnaInfo *retInfo);

/*** Free the mother string in the cdnaInfo.  (The info structure itself normally isn't
 * dynamically allocated. */
void wormFreeCdnaInfo(struct wormCdnaInfo *ci);

/*** Return info if any on gene/ORF, or NULL if none exists. wormFreeCdnaInfo retInfo when done.
 */
boolean wormInfoForGene(char *orfName, struct wormCdnaInfo *retInfo);


/*** Get a single worm cDNA sequence. Optionally (if retInfo is non-null) get additional
 * info about the sequence. */
boolean wormCdnaSeq(char *name, struct dnaSeq **retDna, struct wormCdnaInfo *retInfo);

/*** Get rid of any resources used caching or quickly accessing
 * worm cDNA */
void wormCdnaUncache();


/*** Stuff for searching entire database of worm cDNA */

struct wormCdnaIterator
    {
    FILE *faFile;
    };

/*** Set up to search entire database or worm cDNA */
boolean wormSearchAllCdna(struct wormCdnaIterator **retSi);

/*** Free up iterator returned by wormSearchAllCdna() */
void freeWormCdnaIterator(struct wormCdnaIterator **pIt);

/*** Return next sequence in database */
struct dnaSeq *nextWormCdna(struct wormCdnaIterator *it);

/*** Return next sequence and associated info from database. */
boolean nextWormCdnaAndInfo(struct wormCdnaIterator *it, struct dnaSeq **retSeq, struct wormCdnaInfo *retInfo);

char *wormFeaturesDir();
/*** Return the features directory. (Includes trailing slash.) */

char *wormChromDir();
/*** Return the directory with the chromosomes. */

char *wormCdnaDir();
/*** Return directory with cDNA data. */

char *wormSangerDir();
/*** Return directory with Sanger specific gene predictions. */

char *wormGenieDir();
/*** Return directory with Genie specific gene predictions. */

char *wormXenoDir();
/*** Return directory with cross-species alignments. */

/*** See if it looks like a worm gene name - that is
 *   abc-12
 * letters followed by a dash followed by a number. */
boolean wormIsGeneName(char *name);

/*** Check to see if the input is formatted correctly to be
 * an ORF. */
boolean wormIsOrfName(char *in);

/*** Returns list of cosmid.N type ORF names that are known by abc-12 type name. */
struct slName *wormGeneToOrfNames(char *name);

/*** Return first ORF synonym to gene. */
char *wormGeneFirstOrfName(char *geneName);

/*** Look for gene type (unc-12 or something) synonym for cosmid.N name. */
boolean wormGeneForOrf(char *orfName, char *geneNameBuf, int bufSize);

/*** Get the DNA associated with a gene.  Optionally upper case exons. */
boolean getWormGeneDna(char *name, DNA **retDna, boolean upcExons);

/*** Load up entire packed worm genome into memory. */
void wormLoadNt4Genome(struct nt4Seq ***retNt4Seq, int *retNt4Count);

/*** Free up packed worm genome. */
void wormFreeNt4Genome(struct nt4Seq ***pNt4Seq);

/*** Return size of worm chromosome. */
int wormChromSize(char *chrom);

/*** Return part of a worm chromosome. */
DNA *wormChromPart(char *chromId, int start, int size);

/*** Return part of a worm chromosome with exons in upper case. */
DNA *wormChromPartExonsUpper(char *chromId, int start, int size);

/*** Get list of worm chromosome names. */
void wormChromNames(char ***retNames, int *retNameCount);

/*** Return index of worm chromosome. */
int wormChromIx(char *name);

/*** Given ix, return worm chromosome official name. */
char *wormChromForIx(int ix);

/*** This returns a pointer to our official string for the chromosome name.
 * (This allows some routines to do directo pointer comparisons rather
 * than string comparisons.) */
char *wormOfficialChromName(char *name);

/*** Return chromosome position of a gene, ORF,  nameless cluster, or cosmid. Returns
 * FALSE if no such gene/cluster. */
boolean wormGeneRange(char *name, char **retChromId, char *retStrand, int *retStart, int *retEnd);

/*** Chop up a string representation of a range within a chromosome and put the
 * pieces into the return variables. Return FALSE if it isn't formatted right. */
boolean wormParseChromRange(char *in, char **retChromId, int *retStart, int *retEnd);

/*** Check to see if the input is formatted correctly to be
 * a range of a chromosome. */
boolean wormIsChromRange(char *in);

/*** Make sure that we stay inside chromosome. */
void wormClipRangeToChrom(char *chrom, int *pStart, int *pEnd);

/*** Returns true if name is of correct format to be a nameless cluster. */
boolean wormIsNamelessCluster(char *name);

/*** Get DNA associated with nameless cluster */
DNA *wormGetNamelessClusterDna(char *name);


/*** This holds info on where something is in the genome. */
struct wormFeature
    char *chrom;    /*** One of names returned by */
    char *chrom;    /*** One of names returned by */
    char *chrom;    /*** One of names returned by */
    char name[1];   /*** Allocated to fit. */
    char name[1];   /*** Allocated to fit. */
    char name[1];   /*** Allocated to fit. */
    };

/*** Allocate a new feature. */
struct wormFeature *newWormFeature(char *name, char *chrom, int start, int end, char typeByte);

/*** Get info on all cDNAs that overlap the range. */
struct wormFeature *wormCdnasInRange(char *chromId, int start, int end);

/*** Get info on all genes that overlap the range. */
struct wormFeature *wormGenesInRange(char *chromId, int start, int end);

/*** Get info on genes that overlap range in a particular set of gene predictions. */
struct wormFeature *wormSomeGenesInRange(char *chromId, int start, int end, char *gdfDir);

/*** Get info on all cosmids that overlap the range. */
struct wormFeature *wormCosmidsInRange(char *chromId, int start, int end);

/*** Return list of cdna alignments that overlap range. */
struct cdaAli *wormCdaAlisInRange(char *chromId, int start, int end);

FILE *wormOpenGoodAli();
/*** Opens good alignment file and reads signature. 
 * (You can then cdaLoadOne() it.) */

/*** Helps managed fast indexed access to gene predictions. */
struct wormGdfCache
    {
    char **pDir;
    struct snof *snof;
    FILE *file;
    };
extern struct wormGdfCache wormSangerGdfCache;
extern struct wormGdfCache wormGenieGdfCache;

/*** Get the named gdfGene. */
struct gdfGene *wormGetGdfGene(char *name);

/*** Get a single gdfGene of given name. */
struct gdfGene *wormGetSomeGdfGene(char *name, struct wormGdfCache *cache);

/*** Get all gdfGenes that start with a given name. */
struct gdfGene *wormGetGdfGeneList(char *baseName, int baseNameSize);

/*** Get all gdfGenes that start with a given name. */
struct gdfGene *wormGetSomeGdfGeneList(char *baseName, int baseNameSize, struct wormGdfCache *cache);

struct gdfGene *wormGdfGenesInRange(char *chrom, int start, int end, 
    struct wormGdfCache *geneFinder);
/*** Get list of genes in range according to given gene finder. */

/*** Free up resources associated with fast GDF access. */
void wormUncacheGdf();

/*** Uncache some gene prediction set. */
void wormUncacheSomeGdf(struct wormGdfCache *cache);

#endif /*** WORMDNA_H */

