/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** dlist.c - Doubly-linked list routines. */
#include "common.h"
#include "dlist.h"


/*** Return a new doubly linked list. */
struct dlList *newDlList()
{
struct dlList *dl;
AllocVar(dl);
dl->head = (struct dlNode *)(&dl->nullMiddle);
dl->tail = (struct dlNode *)(&dl->head);
return dl;
}

/*** Free up a doubly linked list and it's nodes (but not the node values). */
void freeDlList(struct dlList **pList)
{
struct dlList *list = *pList;
if (list != NULL)
    {
    struct dlNode *node, *next;
    for (node = list->head; node->next != NULL; node = next)
        {
        next = node->next;
        freeMem(node);
        }
    freez(pList);
    }
}

/*** Free all values in doubly linked list and the list itself.  (Just calls
 * freeMem on all values. */
void freeDlListAndVals(struct dlList **pList)
{
struct dlList *list = *pList;
if (list != NULL)
    {
    struct dlNode *node;
    for (node = list->head; node->next != NULL; node = node->next)
        freeMem(node->val);
    freeDlList(pList);
    }
}


void dlInsertBetween(struct dlNode *before, struct dlNode *after, struct dlNode *newNode)
{
before->next = newNode; 
newNode->prev = before; 
newNode->next = after;  
after->prev = newNode; 
}

/*** Add a node to list before anchor member. */
void dlAddBefore(struct dlNode *anchor, struct dlNode *newNode)
{
dlInsertBetween(anchor->prev, anchor, newNode);
}

/*** Add a node to list after anchor member. */
void dlAddAfter(struct dlNode *anchor, struct dlNode *newNode)
{
dlInsertBetween(anchor, anchor->next, newNode);
}

/*** Add a node to head of list. */
void dlAddHead(struct dlList *list, struct dlNode *newNode)
{
struct dlNode *head = list->head;
dlInsertBetween(head->prev, head, newNode);
}

/*** Add a node to tail of list. */
void dlAddTail(struct dlList *list, struct dlNode *newNode)
{
struct dlNode *tail = list->tail;
dlInsertBetween(tail, tail->next, newNode);
}

/*** Create a node containing val and add to list before anchor member. */
struct dlNode *dlAddValBefore(struct dlNode *anchor, void *val)
{
struct dlNode *node = AllocA(struct dlNode);
node->val = val;
dlAddBefore(anchor, node);
return node;
}

/*** Create a node containing val and add to list after anchor member. */
struct dlNode *dlAddValAfter(struct dlNode *anchor, void *val)
{
struct dlNode *node = AllocA(struct dlNode);
node->val = val;
dlAddAfter(anchor, node);
return node;
}

/*** Create a node containing val and add to head of list. */
struct dlNode *dlAddValHead(struct dlList *list, void *val)
{
struct dlNode *node = AllocA(struct dlNode);
node->val = val;
dlAddHead(list, node);
return node;
}

/*** Create a node containing val and add to tail of list. */
struct dlNode *dlAddValTail(struct dlList *list, void *val)
{
struct dlNode *node = AllocA(struct dlNode);
node->val = val;
dlAddTail(list, node);
return node;
}

/*** Removes a node from list. */
void dlRemove(struct dlNode *node)
{
struct dlNode *before = node->prev;
struct dlNode *after = node->next;
before->next = after;
after->prev = before;
}

/*** Removes head from list. */
void dlRemoveHead(struct dlList *list)
{
dlRemove(list->head);
}

/*** Remove tail from list. */
void dlRemoveTail(struct dlList *list)
{
dlRemove(list->tail);
}

/*** Remove first node from list and return it. */
struct dlNode *dlPopHead(struct dlList *list)
{
struct dlNode *node = list->head;
if (node->next == NULL)
    return NULL;
dlRemove(node);
return node;
}

/*** Remove last node from list and return it. */
struct dlNode *dlPopTail(struct dlList *list)
{
struct dlNode *node = list->tail;
if (node->prev == NULL)
    return NULL;
dlRemove(node);
return node;
}

/*** Return length of list. */
int dlCount(struct dlList *list)
{
return slCount(list->head) - 1;
}

/*** Sort a singly linked list with Qsort and a temporary array. 
 * The arguments to the compare function in real, non-void, life
 * are pointers to pointers of the type that is in the val field of 
 * the nodes of the list. */
void dlSort(struct dlList *list, int (*compare )(const void *elem1,  const void *elem2))
{
int len = dlCount(list);

if (len > 1)
    {
    /*** Move val's onto an array, sort, and then put back into list. */
    void **sorter = needLargeMem(len * sizeof(sorter[0]));
    struct dlNode *node;
    int i;

    for (i=0, node = list->head; i<len; ++i, node = node->next)
        sorter[i] = node->val;
    qsort(sorter, len, sizeof(sorter[0]), compare);
    for (i=0, node = list->head; i<len; ++i, node = node->next)
        node->val = sorter[i];
    freeMem(sorter);
    }
}

/*** Return TRUE if list is empty. */
boolean dlEmpty(struct dlList *list)
{
return (list->head->next == NULL);
}

/*** Return slList from dlList. */
void *dlListToSlList(struct dlList *dList)
{
struct slList *list = NULL, *el;
struct dlNode *node;

for (node = dList->tail; node->prev != NULL; node = node->prev)
    {
    el = node->val;
    slAddHead(&list, el);
    }
return list;
}

/*** Move items from b to end of a. */
void dlCat(struct dlList *a, struct dlList *b)
{
struct dlNode *node;
while ((node = dlPopHead(b)) != NULL)
    dlAddTail(a, node);
}

/*** Return node on list if any that has associated val. */
struct dlNode *dlValInList(struct dlList *list, void *val)
{
struct dlNode *node;
for (node = list->head; !dlEnd(node); node = node->next)
    if (node->val == val)
        return node;
return NULL;
}
