/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** ErrAbort.c - our error handler. 
 *
 * This maintains two stacks - a warning message printer
 * stack, and a "abort handler" stack.
 *
 * By default the warnings will go to stderr, and
 * aborts will exit the program.  You can push a
 * function on to the appropriate stack to change
 * this behavior.  The top function on the stack
 * gets called.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "errabort.h"
#include "common.h"

/*** Default error message handler. */
static void defaultVaWarn(char *format, va_list args)
{
if (format != NULL) {
    vfprintf(stderr, format, args);
    fprintf(stderr, "\n");
    }
}

#define maxWarnHandlers 10
static WarnHandler warnArray[maxWarnHandlers] = {defaultVaWarn,};
static int warnIx = 0;

/*** Call top of warning stack to issue warning. */
void vaWarn(char *format, va_list args)
{
warnArray[warnIx](format, args);
}

/*** Issue a warning message. */
void warn(char *format, ...)
{
va_list args;
va_start(args, format);
vaWarn(format, args);
va_end(args);
}

/*** Set abort handler */
void pushWarnHandler(WarnHandler handler)
{
if (warnIx >= maxWarnHandlers-1)
    errAbort("Too many pushWarnHandlers, can only handle %d\n", maxWarnHandlers-1);
warnArray[++warnIx] = handler;
}

/*** Revert to old warn handler. */
void popWarnHandler()
{
if (warnIx <= 0)
    errAbort("Too many popWarnHandlers\n");
--warnIx;
}

/*** Default error handler exits program. */
static void defaultAbort()
{
exit(-1);
}

#define maxAbortHandlers 12
static AbortHandler abortArray[maxAbortHandlers] = {defaultAbort,};
static int abortIx = 0;

/*** Abort without message. */
void noWarnAbort()
{
abortArray[abortIx]();
}

/*** Abort function, with optional (vprintf formatted) error message. */
void vaErrAbort(char *format, va_list args)
{
vaWarn(format, args);
noWarnAbort();
}

/*** Abort function, with optional (printf formatted) error message. */
void errAbort(char *format, ...)
{
va_list args;
va_start(args, format);
vaErrAbort(format, args);
va_end(args);
}

/*** Prints error message from UNIX errno first, then does errAbort. */
void errnoAbort(char *format, ...)
{
char fbuf[512];
va_list args;
va_start(args, format);
sprintf(fbuf, "%s\n%s", strerror(errno), format);
vaErrAbort(fbuf, args);
va_end(args);
}

/*** Set abort handler */
void pushAbortHandler(AbortHandler handler)
{
if (abortIx >= maxAbortHandlers-1)
    errAbort("Too many pushAbortHandlers, can only handle %d\n", maxAbortHandlers-1);
abortArray[++abortIx] = handler;
}

/*** Revert to old abort handler. */
void popAbortHandler()
{
if (abortIx <= 0)
    errAbort("Too many popAbortHandlers\n");
--abortIx;
}

/*** Call the debugger. */
static void debugAbort()
{
fflush(stdout);
assert(FALSE);
defaultAbort();
}

/*** Push abort handler that will invoke debugger. */
void pushDebugAbort()
{
pushAbortHandler(debugAbort);
}
