/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
#include "common.h"
#include "portable.h"
#include "dnautil.h"
#include "dnaseq.h"
#include "fa.h"
#include "linefile.h"

boolean faReadNext(FILE *f, char *defaultName, boolean mustStartWithComment, 
    char **retCommentLine, struct dnaSeq **retSeq)
/*** Read next sequence from .fa file. Return sequence in retSeq.  If retCommentLine is non-null
 * return the '>' line in retCommentLine.   The whole thing returns FALSE at end of file. */
{
char lineBuf[512];
int lineSize;
char *words[1];
int c;
long offset = ftell(f);
size_t dnaSize = 0;
DNA *dna, *sequence, b;
int bogusChars = 0;
char *name = defaultName;

if (name == NULL)
    name = "";
dnaUtilOpen();
if (retCommentLine != NULL)
    *retCommentLine = NULL;
*retSeq = NULL;

/*** Skip first lines until it starts with '>' */
for (;;)
    {
    if(fgets(lineBuf, sizeof(lineBuf), f) == NULL)
        {
        return FALSE;
        }
    lineSize = strlen(lineBuf);
    if (lineBuf[0] == '>')
        {
	if (retCommentLine != NULL)
            *retCommentLine = cloneString(lineBuf);
        offset = ftell(f);
        chopByWhite(lineBuf, words, ArraySize(words));
        name = words[0]+1;
        break;
        }
    else if (!mustStartWithComment)
        {
        fseek(f, offset, SEEK_SET);
        break;
        }
    else
        offset += lineSize;
    }
/*** Count up DNA. */
for (;;)
    {
    c = fgetc(f);
    if (c == EOF || c == '>')
        break;
    if (!isspace(c) && !isdigit(c))
        {
        ++dnaSize;
        }
    }

/*** Allocate DNA and fill it up from file. */
dna = sequence = needHugeMem(dnaSize+1);
fseek(f, offset, SEEK_SET);
for (;;)
    {
    c = fgetc(f);
    if (c == EOF || c == '>')
        break;
    if (!isspace(c) && !isdigit(c))
        {
        if ((b = ntChars[c]) == 0)
	    {
            b = 'n';
	    }
        *dna++ = b;
        }
    }
if (c == '>')
    ungetc(c, f);
*dna = 0;

*retSeq = newDnaSeq(sequence, dnaSize, name);
return TRUE;
}


/*** Read sequence from FA file. Assumes positioned at or before
 * the '>' at start of sequence. */  
struct dnaSeq *faReadOneDnaSeq(FILE *f, char *defaultName, boolean mustStartWithComment)
{
struct dnaSeq *seq;
if (!faReadNext(f, defaultName, mustStartWithComment, NULL, &seq))
    return NULL;
else
    return seq;
}

/*** Convert fa in memory to bioSeq.  Update *pText to point to next
 * record.  Returns NULL when no more sequences left. */
bioSeq *faNextSeqFromMemText(char **pText, boolean isDna)
{
char *name = "";
char *s, *d;
struct dnaSeq *seq;
int size = 0;
char c;
char *filter = (isDna ? ntChars : aaChars);
char *text = *pText;

if (text == NULL)
    return NULL;
dnaUtilOpen();
if (text[0] == '>')
    {
    char *end;
    s = strchr(text, '\n') + 1;
    name = text+1;
    end = skipToSpaces(name);
    if (end != NULL)
        *end = 0;
    }
else
    s = text;
name = cloneString(name);
    
d = text;
for (;;)
    {
    c = *s;
    if (c == 0 || c == '>')
        break;
    ++s;
    if (isspace(c) || isdigit(c))
	continue;
    if ((c = filter[c]) != 0) 
        d[size++] = c;
    else
	{
	if (isDna)
	    d[size++] = 'n';
	else
	    d[size++] = 'X';
	}
    }
d[size] = 0;

/*** Put sequence into our little sequence structure. */
if (size != 0)
    {
    AllocVar(seq);
    seq->name = name;
    seq->dna = text;
    seq->size = size;
    *pText = s;
    return seq;
    }
else
    {
    freeMem(name);
    *pText = NULL;
    return NULL;
    }
}

/*** Convert fa's in memory into list of dnaSeqs. */
bioSeq *faSeqListFromMemText(char *text, boolean isDna)
{
bioSeq *seqList = NULL, *seq;
while ((seq = faNextSeqFromMemText(&text, isDna)) != NULL)
    {
    slAddHead(&seqList, seq);
    }
slReverse(&seqList);
return seqList;
}

/*** Convert fa in memory to bioSeq. */
bioSeq *faSeqFromMemText(char *text, boolean isDna)
{
return faNextSeqFromMemText(&text, isDna);
}

/*** Return a sequence from a .fa file that's been read into
 * a string in memory. This cannabalizes text, which should
 * be allocated with needMem.  This buffer becomes part of
 * the returned dnaSeq, which may be freed normally with
 * freeDnaSeq. */
struct dnaSeq *faFromMemText(char *text)
{
return faNextSeqFromMemText(&text, TRUE);
}

/*** Open fa file and read a single sequence from it. */
struct dnaSeq *faReadSeq(char *fileName, boolean isDna)
{
int maxSize = fileSize(fileName);
int size = 0;
int fd;
DNA *s;
struct dnaSeq *seq;

if (maxSize < 0)
    errAbort("can't open %s", fileName);
s = needHugeMem(maxSize+1);
fd = open(fileName, O_RDONLY);
read(fd, s, maxSize);
close(fd);
s[maxSize] = 0;
return faSeqFromMemText(s, isDna);
}

/*** Open fa file and read a single nucleotide sequence from it. */
struct dnaSeq *faReadDna(char *fileName)
{
return faReadSeq(fileName, TRUE);
}

/*** Open fa file and read a peptide single sequence from it. */
struct dnaSeq *faReadAa(char *fileName)
{
return faReadSeq(fileName, FALSE);
}

static int faFastBufSize = 0;
static DNA *faFastBuf;

/*** Make faFastBuf bigger. */
static void expandFaFastBuf(int bufPos)
{
if (faFastBufSize == 0)
    {
    faFastBufSize = 64 * 1024;
    faFastBuf = needHugeMem(faFastBufSize);
    }
else
    {
    DNA *newBuf;
    int newBufSize = faFastBufSize + faFastBufSize;
    newBuf = needHugeMem(newBufSize);
    memcpy(newBuf, faFastBuf, bufPos);
    freeMem(faFastBuf);
    faFastBuf = newBuf;
    faFastBufSize = newBufSize;
    }
}

/*** Free up buffers used in fa fast and speedreading. */
void faFreeFastBuf()
{
freez(&faFastBuf);
faFastBufSize = 0;
}

/*** Read in next FA entry as fast as we can. Return FALSE at EOF. 
 * The returned DNA and name will be overwritten by the next call
 * to this function. */
boolean faFastReadNext(FILE *f, DNA **retDna, int *retSize, char **retName)
{
int c;
int bufIx = 0;
static char name[256];
int nameIx = 0;
boolean gotSpace = FALSE;

/*** Seek to next '\n' and save first word as name. */
dnaUtilOpen();
name[0] = 0;
for (;;)
    {
    if ((c = fgetc(f)) == EOF)
        {
        *retDna = NULL;
        *retSize = 0;
        *retName = NULL;
        return FALSE;
        }
    if (!gotSpace && nameIx < ArraySize(name)-1)
        {
        if (isspace(c))
            gotSpace = TRUE;
        else if (c != '>')
            {
            name[nameIx++] = c;
            }
        }
    if (c == '\n')
        break;
    }
name[nameIx] = 0;
/*** Read until next '>' */
for (;;)
    {
    c = fgetc(f);
    if (isspace(c) || isdigit(c))
        continue;
    if (c == EOF || c == '>')
        c = 0;
    else
	{
	c = ntChars[c];
	if (c == 0) c = 'n';
	}
    if (bufIx >= faFastBufSize)
	expandFaFastBuf(bufIx);
    faFastBuf[bufIx++] = c;
    if (c == 0)
        {
        *retDna = faFastBuf;
        *retSize = bufIx-1;
        *retName = name;
        return TRUE;
        }
    }
}


/*** Write next sequence to fa file. */
void faWriteNext(FILE *f, char *startLine, DNA *dna, int dnaSize)
{
int dnaLeft = dnaSize;
int lineSize;
if (startLine != NULL)
    fprintf(f, ">%s\n", startLine);

while (dnaLeft > 0)
    {
    lineSize = dnaLeft;
    if (lineSize > 50)
        lineSize = 50;
    mustWrite(f, dna, lineSize);
    fputc('\n', f);
    dna += lineSize;
    dnaLeft -= lineSize;
    }
}

/*** Write out FA file or die trying. */
void faWrite(char *fileName, char *startLine, DNA *dna, int dnaSize)
{
FILE *f = mustOpen(fileName, "w");
faWriteNext(f, startLine, dna, dnaSize);
fclose(f);
}

/*** Write out all sequences in list to file. */
void faWriteAll(char *fileName, bioSeq *seqList)
{
FILE *f = mustOpen(fileName, "w");
bioSeq *seq;

for (seq=seqList; seq != NULL; seq = seq->next)
    faWriteNext(f, seq->name, seq->dna, seq->size);
fclose(f);
}

/*** Read in DNA or Peptide FA record in mixed case.   Allow any upper or lower case
 * letter, or the dash character in. */
boolean faMixedSpeedReadNext(struct lineFile *lf, DNA **retDna, int *retSize, char **retName)
{
char c;
int bufIx = 0;
static char name[256];
int nameIx = 0;
boolean gotSpace = FALSE;
int lineSize, i;
char *line;


/*** Read first line, make sure it starts wiht '>', and read first word
 * as name of sequence. */
name[0] = 0;
if (!lineFileNext(lf, &line, &lineSize))
    {
    *retDna = NULL;
    *retSize = 0;
    return FALSE;
    }
if (line[0] == '>')
    {
    line = firstWordInLine(skipLeadingSpaces(line+1));
    if (line == NULL)
        errAbort("Expecting sequence name after '>' line %d of %s", lf->lineIx, lf->fileName);
    strncpy(name, line, sizeof(name));
    }
else
    {
    errAbort("Expecting '>' line %d of %s", lf->lineIx, lf->fileName);
    }
/*** Read until next '>' */
for (;;)
    {
    if (!lineFileNext(lf, &line, &lineSize))
        break;
    if (line[0] == '>')
        {
	lineFileReuse(lf);
	break;
	}
    if (bufIx + lineSize >= faFastBufSize)
	expandFaFastBuf(bufIx);
    for (i=0; i<lineSize; ++i)
        {
	c = line[i];
	if (isalpha(c) || c == '-')
	    faFastBuf[bufIx++] = c;
	}
    }
if (bufIx >= faFastBufSize)
    expandFaFastBuf(bufIx);
faFastBuf[bufIx] = 0;
*retDna = faFastBuf;
*retSize = bufIx;
*retName = name;
return TRUE;
}

/*** Convert possibly mixed-case protein to upper case.  Also
 * convert any strange characters to 'X'.  Does not change size.
 * of sequence. */
void faToProtein(char *poly, int size)
{
int i;
char c;
dnaUtilOpen();
for (i=0; i<size; ++i)
    {
    if ((c = aaChars[poly[i]]) == 0)
	c = 'X';
    poly[i] = c;
    }
}

/*** Convert possibly mixed-case DNA to lower case.  Also turn
 * any strange characters to 'n'.  Does not change size.
 * of sequence. */
void faToDna(char *poly, int size)
{
int i;
char c;
dnaUtilOpen();
for (i=0; i<size; ++i)
    {
    if ((c = ntChars[poly[i]]) == 0)
	c = 'n';
    poly[i] = c;
    }
}

/*** Read in DNA or Peptide FA record. */
boolean faSomeSpeedReadNext(struct lineFile *lf, DNA **retDna, int *retSize, char **retName, boolean isDna)
{
char *poly;
int size;

dnaUtilOpen();
if (!faMixedSpeedReadNext(lf, retDna, retSize, retName))
    return FALSE;
size = *retSize;
poly = *retDna;
if (isDna)
    faToDna(poly, size);
else
    faToProtein(poly, size);
return TRUE;
}

/*** Read in next peptide FA entry as fast as we can.  */
boolean faPepSpeedReadNext(struct lineFile *lf, DNA **retDna, int *retSize, char **retName)
{
return faSomeSpeedReadNext(lf, retDna, retSize, retName, FALSE);
}

/*** Read in next FA entry as fast as we can. Faster than that old,
 * pokey faFastReadNext. Return FALSE at EOF. 
 * The returned DNA and name will be overwritten by the next call
 * to this function. */
boolean faSpeedReadNext(struct lineFile *lf, DNA **retDna, int *retSize, char **retName)
{
return faSomeSpeedReadNext(lf, retDna, retSize, retName, TRUE);
}

/*** Return list of all sequences in FA file. */
struct dnaSeq *faReadAllSeq(char *fileName, boolean isDna)
{
struct lineFile *lf = lineFileOpen(fileName, FALSE);
struct dnaSeq *seqList = NULL, *seq;
DNA *dna;
char *name;
int size;

while (faSomeSpeedReadNext(lf, &dna, &size, &name, isDna))
    {
    AllocVar(seq);
    seq->name = cloneString(name);
    seq->size = size;
    seq->dna = cloneMem(dna, size+1);
    slAddHead(&seqList, seq);
    }
lineFileClose(&lf);
slReverse(&seqList);
faFreeFastBuf();
return seqList;
}

/*** Return list of all DNA sequences in FA file. */
struct dnaSeq *faReadAllDna(char *fileName)
{
return faReadAllSeq(fileName, TRUE);
}

/*** Return list of all peptide sequences in FA file. */
struct dnaSeq *faReadAllPep(char *fileName)
{
return faReadAllSeq(fileName, FALSE);
}

