/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** Hash.c - implements hashing. */
#include "common.h"
#include "localmem.h"
#include "hash.h"

/*** Returns a CRC value on string. */
bits32 hashCrc(char *string)
{
unsigned char *us = (unsigned char *)string;
unsigned char c;
bits32 shiftAcc = 0;
bits32 addAcc = 0;

while ((c = *us++) != 0)
    {
    shiftAcc <<= 2;
    shiftAcc += c;
    addAcc += c;
    }
return shiftAcc + addAcc;
}

/*** Looks for name in hash table. Returns associated element,
 * if found, or NULL if not. */
struct hashEl *hashLookup(struct hash *hash, char *name)
{
struct hashEl *el = hash->table[hashCrc(name)&hash->mask];
while (el != NULL)
    {
    if (strcmp(el->name, name) == 0)
        break;
    el = el->next;
    }
return el;
}

/*** Add new element to hash table. */
struct hashEl *hashAdd(struct hash *hash, char *name, void *val)
{
struct hashEl *el = lmAlloc(hash->lm, sizeof(*el));
int hashVal = (hashCrc(name)&hash->mask);
int len = strlen(name);
el->name = lmAlloc(hash->lm, len+1);
strcpy(el->name, name);
el->val = val;
el->next = hash->table[hashVal];
hash->table[hashVal] = el;
return el;
}

/*** Remove item of the given name from hash table. 
 * Returns value of removed item. */
void *hashRemove(struct hash *hash, char *name)
{
struct hashEl *hel;
void *ret;
struct hashEl **pBucket = &hash->table[hashCrc(name)&hash->mask];
for (hel = *pBucket; hel != NULL; hel = hel->next)
    if (sameString(hel->name, name))
        break;
if (hel == NULL)
    {
    warn("Trying to remove non-existant %s from hash", name);
    return NULL;
    }
ret = hel->val;
slRemoveEl(pBucket, hel);
return ret;
}

/*** Add new element to hash table. Squawk and die if not unique */
struct hashEl *hashAddUnique(struct hash *hash, char *name, void *val)
{
if (hashLookup(hash, name) != NULL)
    errAbort("%s duplicated, aborting", name);
return hashAdd(hash, name, val);
}

/*** Add new element to hash table.  Save the name of the element, which is now
 * allocated in the hash table, to *saveName.  A typical usage would be:
 *    AllocVar(el);
 *    hashAddSaveName(hash, name, el, &el->name);
 */
struct hashEl *hashAddSaveName(struct hash *hash, char *name, void *val, char **saveName)
{
struct hashEl *hel = hashAdd(hash, name, val);
*saveName = hel->name;
return hel;
}

/*** If element in hash already return it, otherwise add it
 * and return it. */
struct hashEl *hashStore(struct hash *hash, char *name)
{
struct hashEl *hel;
if ((hel = hashLookup(hash, name)) != NULL)
    return hel;
return hashAdd(hash, name, NULL);
}

/*** If element in hash already return it, otherwise add it
 * and return it. */
char  *hashStoreName(struct hash *hash, char *name)
{
struct hashEl *hel;
if (name == NULL)
    return NULL;
if ((hel = hashLookup(hash, name)) != NULL)
    return hel->name;
return hashAdd(hash, name, NULL)->name;
}

/*** Lookup name in hash and return val.  Abort if not found. */
void *hashMustFindVal(struct hash *hash, char *name)
{
struct hashEl *hel = hashLookup(hash, name);
if (hel == NULL)
    errAbort("%s not found", name);
return hel->val;
}

/*** Look up name in hash and return val or NULL if not found. */
void *hashFindVal(struct hash *hash, char *name)
{
struct hashEl *hel = hashLookup(hash, name);
if (hel == NULL)
    return NULL;
return hel->val;
}

/*** Return name as stored in hash table (in hel->name). 
 * Abort if not found. */
char *hashMustFindName(struct hash *hash, char *name)
{
struct hashEl *hel = hashLookup(hash, name);
if (hel == NULL)
    errAbort("%s not found", name);
return hel->name;
}

/*** Returns new hash table. */
struct hash *newHash(int powerOfTwoSize)
{
struct hash *hash = needMem(sizeof(*hash));
int memBlockPower = 16;
if (powerOfTwoSize == 0)
    powerOfTwoSize = 12;
assert(powerOfTwoSize <= 24 && powerOfTwoSize > 0);
hash->powerOfTwoSize = powerOfTwoSize;
hash->size = (1<<powerOfTwoSize);
/*** Make size of memory block for allocator vary between
 * 256 bytes and 64k depending on size of table. */
if (powerOfTwoSize < 8)
    memBlockPower = 8;
else if (powerOfTwoSize < 16)
    memBlockPower = powerOfTwoSize;
hash->lm = lmInit(1<<memBlockPower);
hash->mask = hash->size-1;
hash->table = lmAlloc(hash->lm, sizeof(struct hashEl *) * hash->size);
return hash;
}

/*** Apply func to every element of hash with hashEl as parameter. */
void hashTraverseEls(struct hash *hash, void (*func)(struct hashEl *hel))
{
int i;
struct hashEl *hel;
for (i=0; i<hash->size; ++i)
    {
    for (hel = hash->table[i]; hel != NULL; hel = hel->next)
	func(hel);
    }
}

/*** Apply func to every element of hash with hashEl->val as parameter. */
void hashTraverseVals(struct hash *hash, void (*func)(void *val))
{
int i;
struct hashEl *hel;
for (i=0; i<hash->size; ++i)
    {
    for (hel = hash->table[i]; hel != NULL; hel = hel->next)
	func(hel->val);
    }
}

/*** Return a list of all elements of hash.   Free return with hashElFreeList. */
struct hashEl *hashElListHash(struct hash *hash)
{
int i;
struct hashEl *hel, *dupe, *list = NULL;
for (i=0; i<hash->size; ++i)
    {
    for (hel = hash->table[i]; hel != NULL; hel = hel->next)
	{
	dupe = cloneMem(hel, sizeof(*hel));
	slAddHead(&list, hel);
	}
    }
return list;
}


/*** Free hash el list returned from hashListAll.  (Don't use
 * this internally.) */
void hashElFree(struct hashEl **pEl)
{
freez(pEl);
}

/*** Free hash el list returned from hashListAll.  (Don't use
 * this internally. */
void hashElFreeList(struct hashEl **pList)
{
slFreeList(pList);
}


/*** Free up hash table. */
void freeHash(struct hash **pHash)
{
struct hash *hash = *pHash;
if (hash == NULL)
    return;
lmCleanup(&hash->lm);
freez(pHash);
}

/*** Free up hash table and all values associated with it.
 * (Just calls freeMem on each hel->val) */
void freeHashAndVals(struct hash **pHash)
{
struct hash *hash;
if ((hash = *pHash) != NULL)
    {
    hashTraverseVals(hash, freeMem);
    freeHash(pHash);
    }
}

