/*******************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/*** lineFile - stuff to rapidly read text files and parse them into
 * lines. */

#include "common.h"
#include <fcntl.h>
#include "linefile.h"

/*** Wrap a line file around an open'd file. */
struct lineFile *lineFileAttatch(char *fileName, bool zTerm, int fd)
{
struct lineFile *lf;
AllocVar(lf);
lf->fileName = cloneString(fileName);
lf->fd = fd;
lf->bufSize = 64*1024;
lf->zTerm = zTerm;
lf->buf = needMem(lf->bufSize+1);
return lf;
}

/*** Wrap a line file around stdin. */
struct lineFile *lineFileStdin(bool zTerm)
{
return lineFileAttatch("stdin", zTerm, fileno(stdin));
}

/*** Try and open up a lineFile. */
struct lineFile *lineFileMayOpen(char *fileName, bool zTerm)
{
int fd;

if (sameString(fileName, "stdin"))
    return lineFileStdin(zTerm);
fd = open(fileName, O_RDONLY);
if (fd == -1)
    return NULL;
return lineFileAttatch(fileName, zTerm, fd);
}

/*** Open up a lineFile or die trying. */
struct lineFile *lineFileOpen(char *fileName, bool zTerm)
{
struct lineFile *lf = lineFileMayOpen(fileName, zTerm);
if (lf == NULL)
    errAbort("Couldn't open %s", fileName);
return lf;
}

/*** Reuse current line. */
void lineFileReuse(struct lineFile *lf)
{
lf->reuse = TRUE;
}

/*** Fetch next line from file. */
boolean lineFileNext(struct lineFile *lf, char **retStart, int *retSize)
{
char *buf = lf->buf;
int bytesInBuf = lf->bytesInBuf;
int endIx;
boolean gotLf = FALSE;
int newStart;

if (lf->reuse)
    {
    lf->reuse = FALSE;
    if (retSize != NULL)
	*retSize = lf->lineEnd - lf->lineStart;
    *retStart = buf + lf->lineStart;
    return TRUE;
    }
/*** Find next end of line in buffer. */
for (endIx = lf->lineEnd; endIx < bytesInBuf; ++endIx)
    {
    if (buf[endIx] == '\n')
	{
	gotLf = TRUE;
	endIx += 1;
	break;
	}
    }
/*** If not in buffer read in a new buffer's worth. */
if (!gotLf)
    {
    int oldEnd = lf->lineEnd;
    int sizeLeft = bytesInBuf - oldEnd;
    int bufSize = lf->bufSize;
    int readSize = bufSize - sizeLeft;

    memmove(buf, buf+oldEnd, sizeLeft);
    lf->bufOffsetInFile += oldEnd;
    readSize = read(lf->fd, buf+sizeLeft, readSize);
    if (readSize + sizeLeft <= 0)
	{
	lf->bytesInBuf = lf->lineStart = lf->lineEnd = 0;
	return FALSE;
	}
    bytesInBuf = lf->bytesInBuf = readSize + sizeLeft;
    lf->lineEnd = 0;

    /*** Look for next end of line.  If can't find it
     * squawk and die. */
    for (endIx = sizeLeft; endIx <bytesInBuf; ++endIx)
	{
	if (buf[endIx] == '\n')
	    {
	    endIx += 1;
	    gotLf = TRUE;
	    break;
	    }
	}
    if (!gotLf && bytesInBuf == lf->bufSize)
        {
	errAbort("Line too long (more than %d chars) line %d of %s",
	    lf->bufSize, lf->lineIx+1, lf->fileName);
	}
		
    }

if (lf->zTerm)
    {
    buf[endIx-1] = 0;
    //uglyf("%03d: %s\n", lf->lineIx, buf+lf->lineEnd);
    }
lf->lineStart = newStart = lf->lineEnd;
lf->lineEnd = endIx;
++lf->lineIx;
if (retSize != NULL)
    *retSize = endIx - newStart;
*retStart = buf + newStart;
return TRUE;
}

/*** Fetch next line from file.  Squawk and die if it's not there. */
void lineFileNeedNext(struct lineFile *lf, char **retStart, int *retSize)
{
if (!lineFileNext(lf, retStart, retSize))
    errAbort("Unexpected end of file in %s", lf->fileName);
}

/*** Close up a line file. */
void lineFileClose(struct lineFile **pLf)
{
struct lineFile *lf;
if ((lf = *pLf) != NULL)
    {
    if (lf->fd > 0 && lf->fd != fileno(stdin))
	close(lf->fd);
    freeMem(lf->fileName);
    freeMem(lf->buf);
    freez(pLf);
    }
}

/*** Check line has right number of words. */
void lineFileExpectWords(struct lineFile *lf, int expecting, int got)
{
if (expecting != got)
    errAbort("Expecting %d words line %d of %s got %d", 
	    expecting, lf->lineIx, lf->fileName, got);
}

/*** Return next non-blank line that doesn't start with '#' chopped into words. */
int lineFileChopNext(struct lineFile *lf, char *words[], int maxWords)
{
int lineSize, wordCount;
char *line;

while (lineFileNext(lf, &line, &lineSize))
    {
    if (line[0] == '#')
        continue;
    wordCount = chopByWhite(line, words, maxWords);
    if (wordCount != 0)
        return wordCount;
    }
return 0;
}

/*** Return next non-blank line that doesn't start with '#' chopped into words.
 * Returns FALSE at EOF.  Aborts on error. */
boolean lineFileNextRow(struct lineFile *lf, char *words[], int wordCount)
{
int wordsRead;
wordsRead = lineFileChopNext(lf, words, wordCount);
if (wordsRead == 0)
    return FALSE;
if (wordsRead < wordCount)
    lineFileExpectWords(lf, wordCount, wordsRead);
return TRUE;
}

/*** Make sure that words[wordIx] is an ascii integer, and return
 * binary representation of it. */
int lineFileNeedNum(struct lineFile *lf, char *words[], int wordIx)
{
char *ascii = words[wordIx];
char c = ascii[0];
if (c != '-' && !isdigit(c))
    errAbort("Expecting number field %d line %d of %s, got %s", 
    	wordIx+1, lf->lineIx, lf->fileName, ascii);
return atoi(ascii);
}

/*** Skip a number of lines. */
void lineFileSkip(struct lineFile *lf, int lineCount)
{
int i, lineSize;
char *line;

for (i=0; i<lineCount; ++i)
    {
    if (!lineFileNext(lf, &line, &lineSize))
        errAbort("Premature end of file in %s", lf->fileName);
    }
}
